"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
import re
import json
from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.models.dbsetting import DBSetting


def get_payoff_forms(selenium):
    form_ids = [
        s.get_attribute('id')
        for s in selenium.find_elements_by_tag_name('form')
    ]
    res = {}
    for f_type in ['increase', 'onetime']:
        vals = {}
        id_re = re.compile(r'^payoff_' + f_type + r'_frm_(\d+)$')
        for frmid in form_ids:
            m = id_re.match(frmid)
            if m is None:
                continue
            _id = int(m.group(1))

            vals[_id] = {
                'enabled': selenium.find_element_by_id(
                    'payoff_%s_frm_%s_enable' % (f_type, _id)
                ).is_selected(),
                'date': selenium.find_element_by_id(
                    'payoff_%s_frm_%s_date' % (f_type, _id)
                ).get_attribute('value'),
                'amount': selenium.find_element_by_id(
                    'payoff_%s_frm_%s_amt' % (f_type, _id)
                ).get_attribute('value')
            }
        res[f_type] = vals
    return res


def set_payoff_form(selenium, f_type, idx, enabled, date_s, amt):
    is_enabled = selenium.find_element_by_id(
        'payoff_%s_frm_%s_enable' % (f_type, idx)).is_selected()
    if is_enabled != enabled:
        selenium.find_element_by_id(
            'payoff_%s_frm_%s_enable' % (f_type, idx)).click()
        assert selenium.find_element_by_id(
            'payoff_%s_frm_%s_enable' % (f_type, idx)).is_selected() == enabled
    selenium.find_element_by_id(
        'payoff_%s_frm_%s_date' % (f_type, idx)).clear()
    selenium.find_element_by_id(
        'payoff_%s_frm_%s_date' % (f_type, idx)).send_keys(date_s)
    selenium.find_element_by_id(
        'payoff_%s_frm_%s_amt' % (f_type, idx)).clear()
    selenium.find_element_by_id(
        'payoff_%s_frm_%s_amt' % (f_type, idx)).send_keys(amt)


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestCreditPayoffs(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/accounts/credit-payoff')

    def test_heading(self, selenium):
        heading = selenium.find_element_by_class_name('navbar-brand')
        assert heading.text == 'Credit Card Payoffs - BiweeklyBudget'

    def test_nav_menu(self, selenium):
        ul = selenium.find_element_by_id('side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_notifications(self, selenium):
        div = selenium.find_element_by_id('notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'


@pytest.mark.acceptance
@pytest.mark.usefixtures('testflask')
@pytest.mark.incremental
class TestNoSettings(AcceptanceHelper):

    def test_00_verify_db(self, testdb):
        b = testdb.query(DBSetting).get('credit-payoff')
        assert b is None

    def test_01_no_config(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        assert selenium.find_element_by_id(
            'payoff_frm_min_pymt').get_attribute('value') == '144.98'
        form_ids = [
            s.get_attribute('id')
            for s in selenium.find_elements_by_tag_name('form')
        ]
        assert sorted(form_ids) == [
            'min_payment_frm',
            'payoff_increase_frm_1',
            'payoff_onetime_frm_1'
        ]
        assert get_payoff_forms(selenium) == {
            'increase': {
                1: {
                    'enabled': False,
                    'date': '',
                    'amount': ''
                }
            },
            'onetime': {
                1: {
                    'enabled': False,
                    'date': '',
                    'amount': ''
                }
            }
        }

    def test_01_highest_balance_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element_by_id('table_HighestBalanceFirstMethod')
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)',
                '2.3 years', '$963.00', '$10.94'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)',
                '4.6 years', '$6,956.35', '$1,457.70'
            ],
            ['Totals', '4.6 years', '$7,919.34', '$1,468.63']
        ]

    def test_02_highest_interest_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element_by_id(
            'table_HighestInterestRateFirstMethod'
        )
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)',
                '2.3 years', '$963.00', '$10.94'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)',
                '4.6 years', '$6,956.35', '$1,457.70'
            ],
            ['Totals', '4.6 years', '$7,919.34', '$1,468.63']
        ]

    def test_03_lowest_balance_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element_by_id('table_LowestBalanceFirstMethod')
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)',
                '1.8 years', '$960.92', '$8.86'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)',
                '4.7 years', '$6,988.24', '$1,489.59'
            ],
            ['Totals', '4.7 years', '$7,949.15', '$1,498.44']
        ]

    def test_04_lowest_interest_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element_by_id(
            'table_LowestInterestRateFirstMethod'
        )
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)',
                '1.8 years', '$960.92', '$8.86'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)',
                '4.7 years', '$6,988.24', '$1,489.59'
            ],
            ['Totals', '4.7 years', '$7,949.15', '$1,498.44']
        ]

    def test_05_min_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element_by_id('table_MinPaymentMethod')
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)',
                '2.3 years', '$963.00', '$10.94'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)',
                '13.5 years', '$8,664.86', '$3,166.21'
            ],
            ['Totals', '13.5 years', '$9,627.86', '$3,177.15']
        ]


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
class TestSettings(AcceptanceHelper):

    def test_00_verify_db(self, testdb):
        b = testdb.query(DBSetting).get('credit-payoff')
        assert b is None

    def test_01_no_config(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        form_ids = [
            s.get_attribute('id')
            for s in selenium.find_elements_by_tag_name('form')
        ]
        assert sorted(form_ids) == [
            'min_payment_frm',
            'payoff_increase_frm_1',
            'payoff_onetime_frm_1'
        ]
        assert get_payoff_forms(selenium) == {
            'increase': {
                1: {
                    'enabled': False,
                    'date': '',
                    'amount': ''
                }
            },
            'onetime': {
                1: {
                    'enabled': False,
                    'date': '',
                    'amount': ''
                }
            }
        }

    def test_02_input_and_save(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        assert 'disabled' in selenium.find_element_by_id(
            'btn_recalc_payoffs').get_attribute('class')
        set_payoff_form(selenium, 'increase', 1, True, '2017-05-14', '160.23')
        assert 'disabled' not in selenium.find_element_by_id(
            'btn_recalc_payoffs').get_attribute('class')
        selenium.find_element_by_link_text('(add another increase)').click()
        set_payoff_form(selenium, 'increase', 2, False, '2017-06-27', '-56.78')
        set_payoff_form(selenium, 'onetime', 1, False, '2017-07-21', '98.76')
        selenium.find_element_by_link_text(
            '(add another one-time additional payment)').click()
        set_payoff_form(selenium, 'onetime', 2, True, '2017-07-04', '54.32')
        selenium.find_element_by_id('btn_recalc_payoffs').click()
        assert len(selenium.find_elements_by_class_name('formfeedback')) == 0

    def test_03_verify_db(self, testdb):
        b = testdb.query(DBSetting).get('credit-payoff')
        assert b is not None
        assert b.value == json.dumps({
            'increases': [
                {
                    'enabled': True,
                    'date': '2017-05-14',
                    'amount': '160.23'
                },
                {
                    'enabled': False,
                    'date': '2017-06-27',
                    'amount': '-56.78'
                }
            ],
            'onetimes': [
                {
                    'enabled': True,
                    'date': '2017-07-04',
                    'amount': '54.32'
                },
                {
                    'enabled': False,
                    'date': '2017-07-21',
                    'amount': '98.76'
                }
            ]
        }, sort_keys=True)

    def test_04_form_fill(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        form_ids = [
            s.get_attribute('id')
            for s in selenium.find_elements_by_tag_name('form')
        ]
        assert sorted(form_ids) == [
            'min_payment_frm',
            'payoff_increase_frm_1',
            'payoff_increase_frm_2',
            'payoff_onetime_frm_1',
            'payoff_onetime_frm_2'
        ]
        assert get_payoff_forms(selenium) == {
            'increase': {
                1: {
                    'enabled': True,
                    'date': '2017-05-14',
                    'amount': '160.23'
                },
                2: {
                    'enabled': False,
                    'date': '2017-06-27',
                    'amount': '-56.78'
                }
            },
            'onetime': {
                1: {
                    'enabled': True,
                    'date': '2017-07-04',
                    'amount': '54.32'
                },
                2: {
                    'enabled': False,
                    'date': '2017-07-21',
                    'amount': '98.76'
                }
            }
        }

    def test_05_highest_balance_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element_by_id('table_HighestBalanceFirstMethod')
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)',
                '2.3 years', '$963.00', '$10.94'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)',
                '4.0 years', '$6,734.42', '$1,235.77'
            ],
            ['Totals', '4.0 years', '$7,697.42', '$1,246.71']
        ]

    def test_06_highest_interest_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element_by_id(
            'table_HighestInterestRateFirstMethod'
        )
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)',
                '2.3 years', '$963.00', '$10.94'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)',
                '4.0 years', '$6,734.42', '$1,235.77'
            ],
            ['Totals', '4.0 years', '$7,697.42', '$1,246.71']
        ]

    def test_07_lowest_balance_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element_by_id('table_LowestBalanceFirstMethod')
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)',
                '1.3 years', '$958.10', '$6.04'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)',
                '4.1 years', '$6,809.61', '$1,310.96'
            ],
            ['Totals', '4.1 years', '$7,767.71', '$1,317.00']
        ]

    def test_08_lowest_interest_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element_by_id(
            'table_LowestInterestRateFirstMethod'
        )
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)',
                '1.3 years', '$958.10', '$6.04'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)',
                '4.1 years', '$6,809.61', '$1,310.96'
            ],
            ['Totals', '4.1 years', '$7,767.71', '$1,317.00']
        ]

    def test_09_min_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element_by_id('table_MinPaymentMethod')
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)',
                '2.3 years', '$963.00', '$10.94'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)',
                '13.5 years', '$8,664.86', '$3,166.21'
            ],
            ['Totals', '13.5 years', '$9,627.86', '$3,177.15']
        ]

    def test_10_input_and_save_remove_lines(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        selenium.find_element_by_id('rm_increase_2_link').click()
        selenium.find_element_by_id('rm_onetime_1_link').click()
        selenium.find_element_by_id('btn_recalc_payoffs').click()
        assert len(selenium.find_elements_by_class_name('formfeedback')) == 0

    def test_11_verify_db(self, testdb):
        b = testdb.query(DBSetting).get('credit-payoff')
        assert b is not None
        assert b.value == json.dumps({
            'increases': [
                {
                    'enabled': True,
                    'date': '2017-05-14',
                    'amount': '160.23'
                }
            ],
            'onetimes': [
                {
                    'enabled': False,
                    'date': '2017-07-21',
                    'amount': '98.76'
                }
            ]
        }, sort_keys=True)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from flask.views import MethodView
from flask import render_template, jsonify
from decimal import Decimal
import json

from biweeklybudget.flaskapp.app import app
from biweeklybudget.flaskapp.views.formhandlerview import FormHandlerView
from biweeklybudget.models.account import Account, AcctType
from biweeklybudget.db import db_session
from biweeklybudget.interest import (
    INTEREST_CALCULATION_NAMES, MIN_PAYMENT_FORMULA_NAMES
)

logger = logging.getLogger(__name__)


class AccountsView(MethodView):
    """
    Render the GET /accounts view using the ``accounts.html`` template.
    """

    def get(self):
        return render_template(
            'accounts.html',
            bank_accounts=db_session.query(Account).filter(
                Account.acct_type == AcctType.Bank,
                Account.is_active == True).all(),  # noqa
            credit_accounts=db_session.query(Account).filter(
                Account.acct_type == AcctType.Credit,
                Account.is_active == True).all(),  # noqa
            investment_accounts=db_session.query(Account).filter(
                Account.acct_type == AcctType.Investment,
                Account.is_active == True).all(),  # noqa
            interest_class_names=INTEREST_CALCULATION_NAMES.keys(),
            min_pay_class_names=MIN_PAYMENT_FORMULA_NAMES.keys()
        )


class OneAccountView(MethodView):
    """
    Render the /accounts/<int:acct_id> view using the ``account.html``
    template.
    """

    def get(self, acct_id):
        return render_template(
            'accounts.html',
            bank_accounts=db_session.query(Account).filter(
                Account.acct_type == AcctType.Bank,
                Account.is_active == True).all(),  # noqa
            credit_accounts=db_session.query(Account).filter(
                Account.acct_type == AcctType.Credit,
                Account.is_active == True).all(),  # noqa
            investment_accounts=db_session.query(Account).filter(
                Account.acct_type == AcctType.Investment,
                Account.is_active == True).all(),  # noqa
            account_id=acct_id,
            interest_class_names=INTEREST_CALCULATION_NAMES.keys(),
            min_pay_class_names=MIN_PAYMENT_FORMULA_NAMES.keys()
        )


class AccountAjax(MethodView):
    """
    Handle GET /ajax/account/<int:account_id> endpoint.
    """

    def get(self, account_id):
        acct = db_session.query(Account).get(account_id)
        return jsonify(acct.as_dict)


class AccountFormHandler(FormHandlerView):
    """
    Handle POST /forms/account
    """

    def validate(self, data):
        """
        Validate the form data. Return None if it is valid, or else a hash of
        field names to list of error strings for each field.

        :param data: submitted form data
        :type data: dict
        :return: None if no errors, or hash of field name to errors for that
          field
        """
        have_errors = False
        errors = {k: [] for k in data.keys()}

        if data.get('name', '').strip() == '':
            errors['name'].append('Name cannot be empty')
            have_errors = True
        try:
            hasattr(AcctType, data['acct_type'])
        except Exception:
            errors['acct_type'].append('"%s" is not a valid Account Type',
                                       data['acct_type'])
        if data['credit_limit'].strip() != '':
            errors = self._validate_decimal('credit_limit', data, errors)
        if data['apr'].strip() != '':
            errors = self._validate_decimal('apr', data, errors)
        if data['ofxgetter_config_json'].strip() != '':
            try:
                json.loads(data['ofxgetter_config_json'])
            except Exception:
                errors['ofxgetter_config_json'].append('Invalid JSON!')
        if data['prime_rate_margin'].strip() != '':
            errors = self._validate_decimal('prime_rate_margin', data, errors)
        if data['interest_class_name'] not in INTEREST_CALCULATION_NAMES:
            errors['interest_class_name'].append('Invalid interest class')
        if data['min_payment_class_name'] not in MIN_PAYMENT_FORMULA_NAMES:
            errors['min_payment_class_name'].append(
                'Invalid minimum payment class name'
            )
        if have_errors:
            return errors
        return None

    def submit(self, data):
        """
        Handle form submission; create or update models in the DB. Raises an
        Exception for any errors.

        :param data: submitted form data
        :type data: dict
        :return: message describing changes to DB (i.e. link to created record)
        :rtype: str
        """
        new_acct = False
        if 'id' in data and data['id'].strip() != '':
            # updating an existing account
            account = db_session.query(Account).get(int(data['id']))
            if account is None:
                raise RuntimeError("Error: no Account with ID %s" % data['id'])
            action = 'updating Account ' + data['id']
        else:
            new_acct = True
            account = Account()
            action = 'creating new Account'
        account.name = data['name'].strip()
        account.description = self.fix_string(data['description'])
        account.acct_type = getattr(AcctType, data['acct_type'])
        if data['ofx_cat_memo_to_name'] == 'true':
            account.ofx_cat_memo_to_name = True
        else:
            account.ofx_cat_memo_to_name = False
        account.vault_creds_path = self.fix_string(data['vault_creds_path'])
        account.ofxgetter_config_json = self.fix_string(
            data['ofxgetter_config_json']
        )
        if data['negate_ofx_amounts'] == 'true':
            account.negate_ofx_amounts = True
        else:
            account.negate_ofx_amounts = False
        if data['reconcile_trans'] == 'true':
            account.reconcile_trans = True
        else:
            account.reconcile_trans = False
        if account.acct_type == AcctType.Credit:
            if data['credit_limit'].strip() != '':
                account.credit_limit = Decimal(data['credit_limit'])
            else:
                account.credit_limit = None
            if data['apr'].strip() != '':
                account.apr = Decimal(data['apr'])
                if account.apr > Decimal('1'):
                    account.apr = account.apr * Decimal('0.01')
            else:
                account.apr = None
            if data['prime_rate_margin'].strip() != '':
                account.prime_rate_margin = Decimal(data['prime_rate_margin'])
                if account.prime_rate_margin > Decimal('1'):
                    account.prime_rate_margin = account.prime_rate_margin * \
                                                Decimal('0.01')
            else:
                account.prime_rate_margin = None
            account.interest_class_name = data['interest_class_name']
            account.min_payment_class_name = data['min_payment_class_name']
        if data['is_active'] == 'true':
            account.is_active = True
        else:
            account.is_active = False
        logger.info('%s: %s', action, account.as_dict)
        db_session.add(account)
        db_session.commit()
        if new_acct:
            account.set_balance(ledger=Decimal('0'), avail=Decimal('0'))
            db_session.add(account)
            db_session.commit()
        return 'Successfully saved Account %d in database.' % account.id


app.add_url_rule('/accounts', view_func=AccountsView.as_view('accounts_view'))
app.add_url_rule(
    '/ajax/account/<int:account_id>',
    view_func=AccountAjax.as_view('account_ajax')
)
app.add_url_rule(
    '/accounts/<int:acct_id>',
    view_func=OneAccountView.as_view('account_view')
)
app.add_url_rule(
    '/forms/account',
    view_func=AccountFormHandler.as_view('account_form')
)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################

This is mostly based on http://flask.pocoo.org/docs/0.12/patterns/sqlalchemy/
"""

import logging
from sqlalchemy import event, inspect

from biweeklybudget.models.budget_model import Budget
from biweeklybudget.models.transaction import Transaction
from biweeklybudget.utils import fmt_currency

logger = logging.getLogger(__name__)


def handle_trans_amount_change(**kwargs):
    """
    Handle change of :py:attr:`.Transaction.actual_amount` for existing
    instances (``id`` is not None). For new instances, we rely on
    :py:func:`~.handle_new_transaction` called via
    :py:func:`~.handle_before_flush`.

    If the Transaction's :py:attr:`~.Transaction.budget` is a
    :py:class:`~.Budget` with :py:attr:`~.Budget.is_periodic` ``False`` (i.e. a
    standing budget), update the Budget's :py:attr:`~.Budget.current_balance`
    for this transaction.

    See: :py:meth:`sqlalchemy.orm.events.AttributeEvents.set`

    :param kwargs: keyword arguments
    :type kwargs: dict
    """
    tgt = kwargs['target']
    if tgt.id is None:
        logger.debug('got Transaction with id None; skipping')
        return
    if tgt.budget.is_periodic:
        logger.debug('got Transaction with periodic budget; skipping')
        return
    value = float(kwargs['value'])
    oldvalue = float(kwargs['oldvalue'])
    session = inspect(tgt).session
    diff = oldvalue - value
    old_budg_curr = float(tgt.budget.current_balance)
    new_budg = old_budg_curr + diff
    logger.info(
        'Handle Transaction %d against standing budget %d UPDATE; '
        'actual_amount change from %s to %s; update budget current_balance '
        'from %s to %s',
        tgt.id, tgt.budget.id, oldvalue, value, old_budg_curr, new_budg
    )
    tgt.budget.current_balance = new_budg
    session.add(tgt.budget)


def handle_new_transaction(session):
    """
    ``before_flush`` event handler
    (:py:meth:`sqlalchemy.orm.events.SessionEvents.before_flush`)
    on the DB session, to handle creation of *new* Transactions. For updates to
    existing Transactions, we rely on :py:func:`~.handle_trans_amount_change`.

    If the Transaction's :py:attr:`~.Transaction.budget` is a
    :py:class:`~.Budget` with :py:attr:`~.Budget.is_periodic` ``False`` (i.e. a
    standing budget), update the Budget's :py:attr:`~.Budget.current_balance`
    for this transaction.

    :param session: current database session
    :type session: sqlalchemy.orm.session.Session
    """
    updated = 0
    for obj in session.new:
        if not isinstance(obj, Transaction):
            continue
        if obj.budget is not None:
            budg = obj.budget
        else:
            budg = session.query(Budget).get(obj.budget_id)
        if budg.is_periodic:
            continue
        logger.debug(
            'Session has new Transaction referencing standing budget id=%s',
            budg.id
        )
        old_amt = float(budg.current_balance)
        budg.current_balance = old_amt - float(obj.actual_amount)
        logger.info(
            'New transaction (%s) for %s against standing budget id=%s; '
            'update budget current_balance from %s to %s', obj.description,
            fmt_currency(obj.actual_amount), budg.id, fmt_currency(old_amt),
            fmt_currency(budg.current_balance)
        )
        session.add(budg)
        updated += 1
    logger.debug(
        'Done handling new transactions; updated %d standing budgets', updated
    )


def handle_before_flush(session, flush_context, instances):
    """
    Hook into ``before_flush``
    (:py:meth:`sqlalchemy.orm.events.SessionEvents.before_flush`)
    on the DB session, to handle updates that need to be made before persisting
    data. Currently, this method just calls a number of other methods to handle
    specific cases:

    * :py:func:`~.handle_new_transaction`

    :param session: current database session
    :type session: sqlalchemy.orm.session.Session
    :param flush_context: internal SQLAlchemy object
    :type flush_context: sqlalchemy.orm.session.UOWTransaction
    :param instances: deprecated
    """
    logger.debug('handle_before_flush handler')
    handle_new_transaction(session)
    logger.debug('handle_before_flush done')


def init_event_listeners(db_session):
    """
    Initialize/register all SQLAlchemy event listeners.

    See http://docs.sqlalchemy.org/en/latest/orm/events.html

    :param db_session: the Database Session
    :type db_session: sqlalchemy.orm.session.Session
    """
    logger.debug('Setting up DB model event listeners')
    event.listen(
        Transaction.actual_amount,
        'set',
        handle_trans_amount_change,
        active_history=True,
        named=True
    )
    event.listen(
        db_session,
        'before_flush',
        handle_before_flush
    )



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
from datetime import timedelta
from decimal import Decimal
import json

from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.prime_rate import PrimeRateCalculator
from biweeklybudget.models import DBSetting
from biweeklybudget.utils import decode_json_datetime, dtnow


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
class TestPrimeRateCalculator(AcceptanceHelper):

    def test_01_verify_db(self, testdb):
        r = testdb.query(DBSetting).get('prime_rate')
        if r is None:
            return
        testdb.delete(r)
        testdb.flush()
        testdb.commit()

    def test_02_query(self, testdb):
        cls = PrimeRateCalculator(testdb)
        r = cls.prime_rate
        assert r >= Decimal('0.01')
        assert r <= Decimal('0.20')
        dbr = testdb.query(DBSetting).get('prime_rate')
        j = json.loads(dbr.value)
        age = dtnow() - decode_json_datetime(j['date'])
        assert age < timedelta(minutes=1)
        assert Decimal(j['value']) == r

    def test_03_verify_db(self, testdb):
        dbr = testdb.query(DBSetting).get('prime_rate')
        j = json.loads(dbr.value)
        age = dtnow() - decode_json_datetime(j['date'])
        assert age < timedelta(minutes=1)

    def test_04_get_again(self, testdb):
        cls = PrimeRateCalculator(testdb)
        r = cls.prime_rate
        assert r >= Decimal('0.01')
        assert r <= Decimal('0.20')



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
from datetime import datetime, timedelta
from pytz import UTC

from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.tests.sqlhelpers import restore_mysqldump
from biweeklybudget.tests.conftest import engine
from biweeklybudget.models import *
from biweeklybudget.settings import PAY_PERIOD_START_DATE
from biweeklybudget.biweeklypayperiod import BiweeklyPayPeriod


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestIndexNavigation(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url)

    def test_heading(self, selenium):
        heading = selenium.find_element_by_class_name('navbar-brand')
        assert heading.text == 'BiweeklyBudget'

    def test_nav_menu(self, selenium):
        ul = selenium.find_element_by_id('side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_notifications(self, selenium):
        div = selenium.find_element_by_id('notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestIndexAccounts(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url)

    def test_bank_table(self, selenium):
        table = selenium.find_element_by_xpath(
            "//div[@id='panel-bank-accounts']//table"
        )
        assert self.thead2list(table) == [
            'Account', 'Balance', 'Unreconciled', 'Difference'
        ]
        assert self.tbody2textlist(table) == [
            ['BankOne', '$12,789.01 (14 hours ago)', '$0.00', '$12,789.01'],
            ['BankTwoStale', '$100.23 (18 days ago)', '-$333.33', '$433.56']
        ]
        links = []
        tbody = table.find_element_by_tag_name('tbody')
        for tr in tbody.find_elements_by_tag_name('tr'):
            td = tr.find_elements_by_tag_name('td')[0]
            links.append(td.get_attribute('innerHTML'))
        assert links == [
            '<a href="/accounts/1">BankOne</a>',
            '<a href="/accounts/2">BankTwoStale</a>',
        ]

    def test_bank_stale_span(self, selenium):
        tbody = selenium.find_element_by_xpath(
            "//div[@id='panel-bank-accounts']//table/tbody"
        )
        rows = tbody.find_elements_by_tag_name('tr')
        bankTwoStale_bal_td = rows[1].find_elements_by_tag_name('td')[1]
        bal_span = bankTwoStale_bal_td.find_elements_by_tag_name('span')[1]
        assert bal_span.text == '(18 days ago)'
        assert bal_span.get_attribute('class') == 'data_age text-danger'

    def test_credit_table(self, selenium):
        table = selenium.find_element_by_xpath(
            "//div[@id='panel-credit-cards']//table"
        )
        assert self.thead2list(table) == [
            'Account', 'Balance', 'Available', 'Avail - Unrec'
        ]
        assert self.tbody2textlist(table) == [
            ['CreditOne', '-$952.06 (13 hours ago)', '$1,047.94', '$825.72'],
            ['CreditTwo', '-$5,498.65 (a day ago)', '$1.35', '$1.35']
        ]
        links = []
        tbody = table.find_element_by_tag_name('tbody')
        for tr in tbody.find_elements_by_tag_name('tr'):
            td = tr.find_elements_by_tag_name('td')[0]
            links.append(td.get_attribute('innerHTML'))
        assert links == [
            '<a href="/accounts/3">CreditOne</a>',
            '<a href="/accounts/4">CreditTwo</a>',
        ]

    def test_investment_table(self, selenium):
        table = selenium.find_element_by_xpath(
            "//div[@id='panel-investment']//table"
        )
        assert self.thead2list(table) == ['Account', 'Value']
        assert self.tbody2textlist(table) == [
            ['InvestmentOne', '$10,362.91 (13 days ago)']
        ]
        links = []
        tbody = table.find_element_by_tag_name('tbody')
        for tr in tbody.find_elements_by_tag_name('tr'):
            td = tr.find_elements_by_tag_name('td')[0]
            links.append(td.get_attribute('innerHTML'))
        assert links == [
            '<a href="/accounts/5">InvestmentOne</a>'
        ]


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestIndexBudgets(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):  # noqa
        self.baseurl = base_url
        self.get(selenium, base_url + '/')

    def test_budgets_table(self, selenium):
        stable = selenium.find_element_by_id('table-standing-budgets')
        stexts = self.tbody2textlist(stable)
        assert stexts == [
            ['Standing1 (4)', '$1,284.23'],
            ['Standing2 (5)', '$9,482.29']
        ]
        selems = self.tbody2elemlist(stable)
        assert selems[0][0].get_attribute(
            'innerHTML') == '<a href="/budgets/4">' \
                            'Standing1 (4)</a>'


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
@pytest.mark.incremental
class TestIndexPayPeriods(AcceptanceHelper):

    def test_0_clean_db(self, dump_file_path):
        # clean the database; empty schema
        restore_mysqldump(dump_file_path, engine, with_data=False)

    def test_1_add_account(self, testdb):
        a = Account(
            description='First Bank Account',
            name='BankOne',
            ofx_cat_memo_to_name=True,
            ofxgetter_config_json='{"foo": "bar"}',
            vault_creds_path='secret/foo/bar/BankOne',
            acct_type=AcctType.Bank
        )
        testdb.add(a)
        a.set_balance(
            overall_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC),
            ledger=1.0,
            ledger_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC)
        )
        testdb.flush()
        testdb.commit()

    def test_2_add_budgets(self, testdb):
        testdb.add(Budget(
            name='1Income',
            is_periodic=True,
            description='1Income',
            starting_balance=1000.00,
            is_income=True
        ))
        testdb.add(Budget(
            name='2Periodic',
            is_periodic=True,
            description='2Periodic',
            starting_balance=500.00
        ))
        testdb.add(Budget(
            name='3Periodic',
            is_periodic=True,
            description='3Periodic',
            starting_balance=0.00
        ))
        testdb.flush()
        testdb.commit()

    def pay_periods(self, db):
        x = BiweeklyPayPeriod.period_for_date(
            (PAY_PERIOD_START_DATE - timedelta(days=2)), db
        )
        periods = []
        for i in range(0, 10):
            periods.append(x)
            x = x.next
        return periods

    def test_3_add_transactions(self, testdb):
        acct = testdb.query(Account).get(1)
        ibudget = testdb.query(Budget).get(1)
        e1budget = testdb.query(Budget).get(2)
        e2budget = testdb.query(Budget).get(3)
        periods = self.pay_periods(testdb)
        # previous pay period
        ppdate = periods[0].start_date
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=1)),
            actual_amount=100.00,
            budgeted_amount=100.00,
            description='prev income',
            account=acct,
            budget=ibudget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=2)),
            actual_amount=250.00,
            description='prev trans 1',
            account=acct,
            budget=e2budget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=3)),
            actual_amount=600.00,
            budgeted_amount=500.00,
            description='prev trans 2',
            account=acct,
            budget=e1budget
        ))
        ppdate = periods[1].start_date
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=1)),
            actual_amount=1400.00,
            budgeted_amount=100.00,
            description='prev income',
            account=acct,
            budget=ibudget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=2)),
            actual_amount=1850.00,
            description='prev trans 1',
            account=acct,
            budget=e2budget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=3)),
            actual_amount=600.00,
            budgeted_amount=500.00,
            description='prev trans 2',
            account=acct,
            budget=e1budget
        ))
        ppdate = periods[2].start_date
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=1)),
            actual_amount=1400.00,
            budgeted_amount=100.00,
            description='prev income',
            account=acct,
            budget=ibudget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=2)),
            actual_amount=788.00,
            description='prev trans 1',
            account=acct,
            budget=e2budget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=3)),
            actual_amount=600.00,
            budgeted_amount=500.00,
            description='prev trans 2',
            account=acct,
            budget=e1budget
        ))
        ppdate = periods[3].start_date
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=1)),
            actual_amount=1400.00,
            budgeted_amount=100.00,
            description='prev income',
            account=acct,
            budget=ibudget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=2)),
            actual_amount=2.00,
            description='prev trans 1',
            account=acct,
            budget=e2budget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=3)),
            actual_amount=600.00,
            budgeted_amount=500.00,
            description='prev trans 2',
            account=acct,
            budget=e1budget
        ))
        testdb.flush()
        testdb.commit()

    def test_4_confirm_sums(self, testdb):
        periods = self.pay_periods(testdb)
        assert periods[0].overall_sums == {
            'allocated': 750.0,
            'spent': 850.0,
            'income': 1000.0,
            'remaining': 150.0
        }
        assert periods[1].overall_sums == {
            'allocated': 2350.0,
            'spent': 2450.0,
            'income': 1400.0,
            'remaining': -1050.0
        }
        assert periods[2].overall_sums == {
            'allocated': 1288.0,
            'spent': 1388.0,
            'income': 1400.0,
            'remaining': 12.0
        }
        assert periods[3].overall_sums == {
            'allocated': 502.0,
            'spent': 602.0,
            'income': 1400.0,
            'remaining': 798.0
        }

    def test_5_pay_periods_table(self, base_url, selenium, testdb):
        periods = self.pay_periods(testdb)
        self.get(selenium, base_url + '/')
        table = selenium.find_element_by_id('pay-period-table')
        texts = self.tbody2textlist(table)
        elems = self.tbody2elemlist(table)
        expected = [
            [
                periods[1].start_date.strftime('%Y-%m-%d') + ' (current)',
                '$2,350.00',
                '$2,450.00',
                '-$1,050.00'
            ],
            [
                periods[2].start_date.strftime('%Y-%m-%d'),
                '$1,288.00',
                '$1,388.00',
                '$12.00'
            ],
            [
                periods[3].start_date.strftime('%Y-%m-%d'),
                '$502.00',
                '$602.00',
                '$798.00'
            ]
        ]
        for i in range(4, 10):
            expected.append([
                periods[i].start_date.strftime('%Y-%m-%d'),
                '$500.00',
                '$0.00',
                '$500.00'
            ])
        assert texts == expected
        # test links
        links = [x[0].get_attribute('innerHTML') for x in elems]
        expected = []
        for idx, period in enumerate(periods):
            if idx == 0:
                continue
            dstr = period.start_date.strftime('%Y-%m-%d')
            s = '<a href="/payperiod/%s">%s</a>' % (dstr, dstr)
            if idx == 1:
                s += ' <em>(current)</em>'
            expected.append(s)
        assert links == expected
        # test red text for negative dollar amounts
        assert elems[0][3].get_attribute('innerHTML') == '<span ' \
            'class="text-danger">-$1,050.00</span>'
        # test highlighted row for current period
        tbody = table.find_element_by_tag_name('tbody')
        trs = tbody.find_elements_by_tag_name('tr')
        assert trs[0].get_attribute('class') == 'info'



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from biweeklybudget.utils import dtnow
from pytz import utc
from datetime import datetime
from freezegun import freeze_time
import sys

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import patch
else:
    from unittest.mock import patch


class TestDtNow(object):

    @freeze_time('2016-05-13 11:21:32', tz_offset=0)
    @patch('biweeklybudget.utils.settings.BIWEEKLYBUDGET_TEST_TIMESTAMP', None)
    def test_dtnow(self):
        assert dtnow() == datetime(
            2016, 5, 13, 11, 21, 32, tzinfo=utc
        )

    def test_dtnow_test(self):
        assert dtnow() == datetime(
            2017, 7, 28, 6, 24, 44, tzinfo=utc
        )



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

##################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
##################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
##################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
##################################################################################
"""

import biweeklybudget.version as version

import re


class TestVersion(object):

    def test_project_url(self):
        expected = 'https://github.com/jantman/biweeklybudget'
        assert version.PROJECT_URL == expected

    def test_is_semver(self):
        # see:
        # https://github.com/mojombo/semver.org/issues/59#issuecomment-57884619
        semver_ptn = re.compile(
            r'^'
            r'(?P<MAJOR>(?:'
            r'0|(?:[1-9]\d*)'
            r'))'
            r'\.'
            r'(?P<MINOR>(?:'
            r'0|(?:[1-9]\d*)'
            r'))'
            r'\.'
            r'(?P<PATCH>(?:'
            r'0|(?:[1-9]\d*)'
            r'))'
            r'(?:-(?P<prerelease>'
            r'[0-9A-Za-z-]+(\.[0-9A-Za-z-]+)*'
            r'))?'
            r'(?:\+(?P<build>'
            r'[0-9A-Za-z-]+(\.[0-9A-Za-z-]+)*'
            r'))?'
            r'$'
        )
        assert semver_ptn.match(version.VERSION) is not None
